"""
The noisemodel, which is a prediction of x-ray resolution
for an arbitrary shaping time
"""
from dataclasses import dataclass
import numpy as np
import pylab as p
import os
import os.path

import HErmes as he
import hepbasestack as hep

import hjson

from .utils import get_stripname
logger = hep.logger.get_logger(10)

########################################################################

@dataclass
class Constants:
    q      : float = 1.6e-19   # electron charge
    k      : float = 1.38e-23  # Boltzmann constant
    eps    : float = 3.6       # ionization energy of silicon, eV
    Rp     : float = 100e6     # parallel resistance of preamp, 100 MOhm
    gm     : float = 18e-3     # transconductance in FET, 18 ms
    Bita   : float = 1
    Fi     : float = 0.367     # noise form factor
    Fv     : float = 1.15      # noise form factor
    Fvf    : float = 0.226     # noise form factor
    Ctot   : float = 70e-12    #pF

    Fi_g4  : float = 0.45      # noise form factor for gauss order 4
    Fv_g4  : float = 1.02      # noise form factor for gauss order 4
    Fvf_g4 : float = 0.52      # noise form factor for gauss order 4
    
    Fi_tr  : float = 0.83      # noise form factor for trapezoid
    Fv_tr  : float = 1         # noise form factor for trapezoid 
    Fvf_tr : float = 0.69      # noise form factor for trapezoid

    def Rs_denom(self, temp_cels):
        return self.Fv/self.Ctot/self.Ctot/(4*self.k*self.T(temp_cels))-self.Bita/self.gm;

    @property
    def Af_denom(self):
        return self.Ctot/self.Ctot/self.Fvf/2/np.pi 

    # FIXME: While not tecnnically a constant, 
    # we try to keep it as constant as possible
    # I know...    
    def T(self,temp_cels):
        return temp_cels+273

CONSTANTS = Constants()

########################################################################

def serialize_noisemodel(noisemodel,\
                         filename):
    """
    Save the noisemodel to a file, serialize via json

    Args:
        noisemodel (HErmes.fitting.model.Model) : the fitted noisemodel
        filename   (str)                        : full path to file
    """
    nm_data = dict()
    nm_data['I_L']      = noisemodel.I_L
    nm_data['A_f']      = noisemodel.A_f
    nm_data['R_S']      = noisemodel.R_S
    nm_data['chi2/ndf'] = noisemodel.chi2_ndf
    nm_data['npoint']   = len(noisemodel.data)
    nm_data['detid']    = noisemodel.detid     
    nm_data['strip']    = noisemodel.stripname 
    logger.info(f'Serializing noisemodel to {filename}')
    hjson.dump(nm_data, open(filename,'w'))
    return None

########################################################################

def extract_parameters_from_noisemodel(noisemodel,\
                                       method='trapezoid'):
    """
    After fitting the noisemodel, get the detector
    relevant parameters out of the fit parameters
    Multiply with the orders of magnitude to get reasonable units.

    Args:
        noisemodel (HErmes.fiting.Model) : noisemodel after fitting

    Keyword Args:
        method     (str)                 : this describes the functional form of the 
                                           used shaping function. Which can be 
                                           either 'trapezoid' or 'gauss'
                                           This affects the weighting factors for Il, Af and R.
    Returns (dict)                       : the relevant paramters
    """

    factor = (2.355*CONSTANTS.eps*1e-3/CONSTANTS.q)*(2.355*CONSTANTS.eps*1e-3/CONSTANTS.q)

    # use the weighting factors for either trapezoid or Gauss 4th order
    if method == 'trapezoid':
        Fi   = CONSTANTS.Fi  
        Fv   = CONSTANTS.Fv
        Fvf  = CONSTANTS.Fvf
    elif method == 'gauss':
        Fi   = CONSTANTS.Fi_g4
        Fv   = CONSTANTS.Fv_g4
        Fvf  = CONSTANTS.Fvf_g4
    else:
        raise ValueError(f"Can't understand {method}. Has to be either 'gauss' or 'trapezoid'")

    # this is necessary in case we did not use
    # minuit for the fitting, then we do have 
    # a different structure for the error dict
    if isinstance(noisemodel.errors, list):
        tmpdict = dict()
        tmpdict['par00'] = noisemodel.errors[0]
        tmpdict['par10'] = noisemodel.errors[1]
        tmpdict['par20'] = noisemodel.errors[2]
        noisemodel.errors = tmpdict


    p0  = noisemodel.best_fit_params[0]/factor
    ep0 = noisemodel.errors['par00']/factor
    p1  = noisemodel.best_fit_params[1]/factor
    ep1 = noisemodel.errors['par10']/factor
    p2  = noisemodel.best_fit_params[2]/factor
    ep2 = noisemodel.errors['par20']/factor
    
    Ileak = (p0/Fi-4*CONSTANTS.k*CONSTANTS.T(-37)/CONSTANTS.Rp)/2/CONSTANTS.q  # A
    eIleak = ep0/p0*Ileak  # A
    Rs = (p1 /Fv / CONSTANTS.Ctot / CONSTANTS.Ctot / (
                4 * CONSTANTS.k * CONSTANTS.T(-37))) - (CONSTANTS.Bita / CONSTANTS.gm)

    Rs_const = Fv*CONSTANTS.Ctot*CONSTANTS.Ctot*(4*CONSTANTS.k*CONSTANTS.T(-37))
    Rs_const_diff =  -CONSTANTS.Bita/CONSTANTS.gm
    if Rs < 0:
        logger.warn(f'Rs: {Rs}')
        logger.warn(f'p1/Rs_const : {p1/Rs_const - Rs_const_diff}')
        logger.warn(f'.. > 0 {p1/Rs_const > 0}')
        logger.warn(f'Rs_const : {Rs_const > 0}')
        logger.warn(f'p1 : {p1}')
        logger.warn(f'{p1 > 0}')
        logger.warn(f'{p1/factor}')
        logger.warn(f'{CONSTANTS.Fv/CONSTANTS.Ctot/CONSTANTS.Ctot/(4*CONSTANTS.k*CONSTANTS.T(-37))-CONSTANTS.Bita/CONSTANTS.gm}')
        logger.warn(f'{p1 > 0}')
        logger.warn('Rs is < 0!')
    eRs = ep1/p1*Rs  # Ohm

    CTOT  = np.sqrt(p1/Fv/(4*CONSTANTS.k*CONSTANTS.T(-37))/(Rs+CONSTANTS.Bita/CONSTANTS.gm));
    logger.warn(f'Calculated {CTOT}')

    Af = p2/CONSTANTS.Ctot/CONSTANTS.Ctot/Fvf/2/np.pi


    #Af = p2/CTOT/CTOT/CONSTANTS.Fvf#/2/CONSTANTS.pi
    eAf = ep2/p2*Af

    result = {'p0'     : p0   *factor,\
              'ep0'    : ep0  *factor,\
              'p1'     : p1   *factor,\
              'ep1'    : ep1  *factor,\
              'p2'     : p2   *factor,\
              'ep2'    : ep2  *factor,\
              'Ileak'  : Ileak*1e9,\
              'eIleak' : eIleak*1e9,\
              'Rs'     : Rs,\
              'eRs'    : eRs,\
              'Af'     : Af*1e13,\
              'eAf'    : eAf*1e13\
             }
    return result

########################################################################

def preamp_noise_model(tau, IL, Rs, Af):
    """
    Direct representation of the ENC for the preamps

    Aregs:
        tau (ndarray)  : shaping times
        IL (float)    : leakage current
        Rs (float)    : serial resistance
        Af (float)    :
    """
    tau = 1e-6 * tau
    #A_f = A_f*1e13
    #I_L = I_L*1e9
    GAMMA = 1 # Field
    A = CONSTANTS.Fi*(CONSTANTS.q*2*IL + (4*CONSTANTS.k*CONSTANTS.T(-37)*(1/CONSTANTS.Rp)))
    B = 4*CONSTANTS.k*CONSTANTS.T(-37)*CONSTANTS.Fv*CONSTANTS.Ctot*CONSTANTS.Ctot*\
        (Rs + (GAMMA/CONSTANTS.gm))
    C = Af*CONSTANTS.Ctot*CONSTANTS.Ctot*CONSTANTS.Fvf
    #print ((A,B,C, 'Refined nm A,B,C'))
    result = A*tau + B/tau + C*np.ones(len(tau))
    result = np.sqrt(result)*2.355*CONSTANTS.eps*(1/CONSTANTS.q)*1e-3
    return result

########################################################################

def noise_model(xs, par0, par1, par2):
    """
    Noise model as to be fit to the resolution vs shaping times
    plot.
    
    Args:
        xs (ndarray)    : input data
        par0 (float)    : substition, will get mapped to I_L
        par1 (float)    : substitution, will get mapped to R_S
        par2 (float)    : substitution, will get mapped to A_f
    """
    # from root script 
    # sqrt([0]*x*1e-6+[1]/(x*1e-6)+[2])

    # mus -> s
    xs = 1e-6*xs
    result = np.sqrt((par0*xs) + (par1/xs) + (np.ones(len(xs))*par2))
    #result = (par0 * xs) + (par1 / xs) + (np.ones(len(xs)) * par2)
    return result

########################################################################

def construct_error_belt(nm, xs=None):
    """ 
    Get the maximum and minimum values for the noisemodel

    Args:
        nm (HErmes.fitting.Model)   : the fitted noisemodel

    Keyword Args:
        xs (numpy.ndarray)          : x data. If none, take noisemodel.xs
    """
    try:
        eI_L = nm.errors['par00']
    except:
        eI_L = nm.errors['IL0']
    try:
        eR_s = nm.errors['par10']
    except:
        eR_s = nm.errors['Rs0']
    try:
        eA_f = nm.errors['par20']
    except:
        eA_f = nm.errors['Af0']

    minpar0 = nm.best_fit_params[0] - eI_L
    if minpar0 < 0: minpar0 = 0
    minpar1 = nm.best_fit_params[1] - eR_s
    if minpar1 < 0: minpar1 = 0
    minpar2 = nm.best_fit_params[2] - eA_f
    if minpar2 < 0: minpar2 = 0
    minmaxpar0 = (nm.best_fit_params[0] + eI_L,
                  minpar0)
    minmaxpar1 = (nm.best_fit_params[1] + eR_s,
                  minpar1)
    minmaxpar2 = (nm.best_fit_params[2] + eA_f,
                  minpar2)

    if xs is None:
        xs = nm.xs
    currentmaximum = nm(xs, minmaxpar0[1],\
                            minmaxpar1[1],\
                            minmaxpar2[1])
    currentminimum = nm(xs, minmaxpar0[0],\
                            minmaxpar1[0],\
                            minmaxpar2[0])
    for i in range(2):
        for j in range(2):
            for l in range(2):
                thisnoisemodel = nm(xs, minmaxpar0[i],\
                                                 minmaxpar1[i],\
                                                 minmaxpar2[i])
                currentminimum = np.minimum(thisnoisemodel, currentminimum)
                currentmaximum = np.maximum(thisnoisemodel, currentmaximum) 
                thisnoisemodel = nm(xs,  minmaxpar0[j],\
                                                  minmaxpar1[i],\
                                                  minmaxpar2[i]) 
                currentminimum = np.minimum(thisnoisemodel, currentminimum)
                currentmaximum = np.maximum(thisnoisemodel, currentmaximum) 
                thisnoisemodel = nm(xs, minmaxpar0[j],\
                                                 minmaxpar1[j],\
                                                 minmaxpar2[i]) 
                currentminimum = np.minimum(thisnoisemodel, currentminimum)
                currentmaximum = np.maximum(thisnoisemodel, currentmaximum) 
                thisnoisemodel = nm(xs, minmaxpar0[j],\
                                                 minmaxpar1[j],\
                                                 minmaxpar2[j]) 
                currentminimum = np.minimum(thisnoisemodel, currentminimum)
                currentmaximum = np.maximum(thisnoisemodel, currentmaximum) 
    return currentminimum, currentmaximum

########################################################################

def fit_noisemodel(xs, ys, ys_err,  ch, detid,\
                   plotdir='.', fig=None,\
                   use_minuit=True,\
                   use_direct_model=False,\
                   debug_minuit=False,
                   method='trapezoid'):
    """
    Perform the fit of the xray resolutions over peaking time
    with the noisemodel

    Args:
        xs           (ndarray)   : peaking times (microseconds)
        ys           (ndarray)   : x-ray resolutions (fwhm)
        ys_err       (ndarray)   : fwhm associated fit errors
        ch               (int)   : digitizer channel/detector strip
        detid            (int)   : detector id (needed for the filename to save the plot)

    Keyword Args:
        fig (Figure)             : use a pre-exisiting figure instance. If None, create new
        plotdir (str)            : directory to save the plot in
        use_minuit       (bool)  : Use minuit for the minimization (recommended)
        use_direct_model (bool)  : use a slightly different implempentation of the noise
                                   model. The difference is that the 3 physics parameters
                                   I_L, R_S and A_f are fitted directly.
        debug_minuit     (bool)  : pass this parameter to the model. It attaches the
                                   iminuit instance to the model, so it is accessible for
                                   later debugging
        fig (matplotlib.Figure)  : A predefined figure install to use
        method            (str)  : Tell the noisemodel what method was used to obtain the 
                                   data, either a trapezoidal shaper or a gaussian. 
                                   This affects the weights for the physics parameters. 
    Returns:
        tuple (matplotlib.figure, model)               
    """
    # channel noise model fit
    logger.info(f'Performing noise model fit for channel {ch}')
    if use_direct_model:
        logger.info(f'Using direct model, fitting I_L, R_S, A_f...')
        noisemodel = he.fitting.Model(preamp_noise_model)
        noisemodel.startparams = (2e-9,1, 8e-13)
    else:
        noisemodel = he.fitting.Model(noise_model)
        noisemodel.startparams = (5e5, 1e-5,1)

    noisemodel.add_data(ys, xs=xs/1000,\
                        data_errs=ys_err,
                        create_distribution=False)
    if use_direct_model:
        noisemodel.fit_to_data(use_minuit=use_minuit, \
                               debug_minuit=debug_minuit, \
                               errors=(1, 1, 1), \
                               bounds=([1e-10,0,1e-14],[1e-8,20,1e-12]),\
                               limits=((1e-10, 1e-8),\
                                       (0, 20),\
                                       (1e-14, 1e-12))
                                      )


    else:
        noisemodel.fit_to_data(use_minuit=use_minuit,\
                           debug_minuit=debug_minuit,\
                           errors=(1,1,1),\
                           bounds =([1e4,1e-7,0],[1e7,1e-4,100]),\
                           limits=((1e4,1e7), \
                                   #(1e-7,1e-4), \
                                   (1.1450365642091253e-05,1e-4),\
                                   (0,100)))

    logger.info(f'Noisemodel fitted, best fit pars: {noisemodel.best_fit_params}')
    if use_direct_model:
        if isinstance(noisemodel.errors, list):
            tmpdict = dict()
            tmpdict['par00'] = noisemodel.errors[0]
            tmpdict['par10'] = noisemodel.errors[1]
            tmpdict['par20'] = noisemodel.errors[2]
            noisemodel.errors = tmpdict
        I_L = noisemodel.best_fit_params[0]
        R_S = noisemodel.best_fit_params[1]
        A_f = noisemodel.best_fit_params[2]
        print (noisemodel.errors)
        try:
            eI_L = noisemodel.errors['par00']
        except:
            eI_L = noisemodel.errors['IL0']
        try:
            eR_S = noisemodel.errors['par10']
        except:
            eR_S = noisemodel.errors['Rs0']
        try:
            eA_f = noisemodel.errors['par20']
        except:
            eA_f = noisemodel.errors['Af0']
        #
        #eR_S = noisemodel.errors['R_s0']
        #eA_f = noisemodel.errors['A_f0']
        infotext = r"\begin{tabular}{lll}"
        infotext += r"noisemodel fit:&&\\ "
        infotext += r"$I_l$ & {:4.2e} &$\pm$ {:4.2e} [nA]\\".format(I_L, eI_L)
        infotext += r"$A_f$ & {:4.2e} &$\pm$ {:4.2e} [$V^2$]\\".format(A_f, eA_f)
        infotext += r"$R_S$ & {:4.2e} &$\pm$ {:4.2e} [$\Omega$]\\".format(R_S, eR_S)
        infotext += r"$\chi^2/ndf$ &{:4.2f} & \\ ".format(noisemodel.chi2_ndf)
        infotext += r"\end{tabular}"
    else:
        params = extract_parameters_from_noisemodel(noisemodel, method=method)
        logger.info(f'Extracted {params} from noisemodel')
        logger.debug(f'..I_L {params["Ileak"]}')
        logger.debug(f'..A_f {params["Af"]}')
        logger.debug(f'..R_s {params["Rs"]}')
        I_L = params["Ileak"]
        A_f = params["Af"]
        R_S = params["Rs"]

        #print (params['Ileak'])
        # create a textbox with some output
        infotext  = r"\begin{tabular}{lll}"
        infotext += r"noisemodel fit:&&\\ "
        infotext += r"p0 & {:4.2e} &$\pm$ {:4.2e}\\".format(params['p0'], params['ep0'])
        infotext += r"p1 & {:4.2e} &$\pm$ {:4.2e}\\".format(params['p1'], params['ep1'])
        infotext += r"p2 & {:4.2e} &$\pm$ {:4.2e}\\".format(params['p2'], params['ep2'])
        infotext += r"$I_l$ & {:4.2e} &$\pm$ {:4.2e} [nA]\\".format(params['Ileak'], params['eIleak'])
        infotext += r"$A_f$ & {:4.2e} &$\pm$ {:4.2e} $\times 1e-13$ [$V^2$]\\".format(params['Af'], params['eAf'])
        infotext += r"$R_S$ & {:4.2e} &$\pm$ {:4.2e} [$\Omega$]\\".format(params['Rs'], params['eRs'])
        #infotext += r"$C_{tot}$ & {:4.2f} [pF]\\ ".format(peakmod.best_fit_params[4], errdict['fwhm1'])
        infotext += r"$\chi^2/ndf$ &{:4.2f} & \\ ".format(noisemodel.chi2_ndf)
        infotext += r"\end{tabular}"

    if fig is None:
        noisemodel_fig = p.figure()
    else:
        noisemodel_fig = fig
    ax = noisemodel_fig.gca()
    xs_for_plt = np.linspace(min(noisemodel.xs), max(noisemodel.xs), 10000)
    ax.plot(xs_for_plt, noisemodel(xs_for_plt, *noisemodel.best_fit_params),\
            color='r', lw=1.2, zorder=1)
    minmodel, maxmodel = construct_error_belt(noisemodel, xs=xs_for_plt)
    ax.fill_between(xs_for_plt, minmodel, y2=maxmodel,\
                    color='r', alpha=0.4, zorder=1)
    ax.errorbar(xs/1000, ys, yerr=ys_err, \
                marker='.', mfc='k', mec='k', ms=1.2, \
                #fmt='none',\
                fmt='.', \
                ecolor='k',\
                lw=1.2,\
                zorder=2,\
               )
    ax = hep.visual.adjust_minor_ticks(ax, which='both')
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_ylim(bottom=1)
    ax.grid(which='minor', color='gray', alpha=0.7)
    title = f'det{detid}-{get_stripname(ch)}'
    ax.set_title(title, loc='right')
    ax.set_xlabel('peaking time [$\mu$s]')
    ax.set_ylabel('xray res. (FWHM) [keV]')
    ax.text(0.3, 0.3, infotext, \
            horizontalalignment='center', \
            verticalalignment='center', \
            transform=ax.transAxes, \
            size='xx-small', \
            bbox=dict(facecolor='white', alpha=0.7, edgecolor=None), \
            )
    pngfilename = os.path.join(plotdir, title + '-nmfit.png')
    noisemodel_file = os.path.join(plotdir, title + '-nmfit.dat') 
    noisemodel_fig.savefig(pngfilename)
    logger.warn(f'Saved file {pngfilename}')
    # attach some metainformation
    noisemodel.I_L = I_L
    noisemodel.A_f = A_f
    noisemodel.R_S = R_S
    noisemodel.detid     = detid
    noisemodel.stripname = get_stripname(ch)
    serialize_noisemodel(noisemodel,\
                         noisemodel_file)
    return noisemodel_fig, noisemodel

########################################################################

def enc2_trapezoid(xs, par0, par1, par2):
    """
    Noise model as to be fit to the resolution vs shaping times
    plot.

    Args:
        xs (ndarray)    : input data
        par0 (float)
        par1 (float)
        par2 (float)
    """
    raise NotImplementedError
    # weighting coefficients
    Aw_1 = 2  # series white
    Aw_2 = 1.67  # series parallel
    Aw_3 = 1.37  # series 1/f

    # mus -> s
    xs = 1e-6 * xs
    result = np.sqrt((par0 * xs) + (par1 / xs) + (np.ones(len(xs)) * par2))
    # fit C_in2, A_f and I_0 directly
    # coefficients
    a_0 = 0.5

    # u = 1/np.sqrt(xs)
    # transform into u space
    # result = np.sqrt(par0)*u**2 + np.sqrt(par1)* u + np.sqrt(par2)
    return result


########################################################################

def enc2_semigauss4(xs, par0, par1, par2):
    """
    Noise model as to be fit to the resolution vs shaping times
    plot.

    Args:
        xs (ndarray)    : input data
        par0 (float)
        par1 (float)
        par2 (float)
    """
    raise NotImplementedError
    # from root script
    # sqrt([0]*x*1e-6+[1]/(x*1e-6)+[2])

    # mus -> s
    xs = 1e-6 * xs
    result = np.sqrt((par0 * xs) + (par1 / xs) + (np.ones(len(xs)) * par2))
    # u = 1/np.sqrt(xs)
    # transform into u space
    # result = np.sqrt(par0)*u**2 + np.sqrt(par1)* u + np.sqrt(par2)
    return result




"""
Specific to the GAPS experiment
Prediction of the energy resolution with the custom ASIC
from the calibration with discrete preamps.
"""

import numpy as np
import hepbasestack as hep
from dataclasses import dataclass
from .noisemodel import Constants

@dataclass
class ASICConstants : 
    S_w   : float = 0.54*1e-18 #V2/Hz
    A_f   : float = 7.3*1e-14 #V2
    I_eff : float = 2.5*1e-9 # A
    # RC CR2 coefficients
    F_i   : float = 0.64 
    F_nu  : float = 0.853
    F_nuf : float = 0.543


CONSTANTS  = Constants()
ACONSTANTS = ASICConstants()

def enc2_asic(tau, I_L,  A_f, C=70*1e-12):
    """
    ASIC prediction from fitted noisemodel
    values from the preamp measurement.

    Args:
        tau (iterable) : xs - the shaping/peaking time
        I_L (float)    : fitted leakage current from the preamp 
                         measurement
        A_f (float)    : 

    Keyword Args:
        C (float)      : fitted capacity**2 from the preamp
                         measurementi - looks like this is always
                         70 pF
        
    """
    # FIXME logging
    if A_f < 0.74*1e-13: # this value comes from Mengjiao
        print (f"WARN: A_f too small {A_f}, will use 0.74!")
        A_f = 0.74*1e-13
    tau = 1e-6*tau
    C_2 = C**2
    A =  2*CONSTANTS.q*(I_L + ACONSTANTS.I_eff)*tau*ACONSTANTS.F_i
    B =  ACONSTANTS.S_w*(C_2)*ACONSTANTS.F_nu*(1/tau)
    C = 2*np.pi*A_f*ACONSTANTS.F_nuf*(C_2)*np.ones(len(tau))
    enc2 = A+B+C
    #print ((A[0],B[0], C[0], 'A,B,C'))
    return np.sqrt(enc2)*2.355*CONSTANTS.eps*(1/CONSTANTS.q)*1e-3
    #return enc2

####################################################################################

def asic_projection_plot(tau, asic,\
                         stripname,\
                         detid,\
                         fig=None,\
                         loglog=False):
    """
    Produce a plot with asic prediction vs shaping time

    Args:
        tau      (numpy.ndarray) : shaping times, x-values
        asic     (numpy.ndarray) : asic prediction, y-values
        stripname          (str) : strip identifier A-H on GAPS Si(Li) detector
        detid              (int) : detector id.
    Keyword Args:
        fig      (matplotlib.figure.Figure) : a pre-initialized figure instance
                                              to plot the data in. If None, new
                                              one gets created.
        loglog                        (bool): create a log-log plot(default False)

    """

    title = f'det {detid}'
    label = f'{stripname}'
    if fig is None:
        fig = p.figure(dpi=120,\
                       figsize=hep.layout.FIGSIZE_A4_LANDSCAPE_HALF_HEIGHT)

    ax = fig.gca()
    #ax.plot(nm.xs, asic, label=label, lw=1.2)
    ax.plot(tau, asic, label=label, lw=1.2)
    ax.set_xlim(left=0.49, right=2)
    ax.set_ylim(top=6)
    ax.set_title(title, loc='right')
    # ax.legend()
    # hep.visual.adjust_minor_ticks(ax, which='both')
    ax.grid(which='minor', color='gray', alpha=0.7)
    if loglog:
        ax.set_xscale('log')
        ax.set_yscale('log')
    ax.set_ylabel('proj. ASIC FWHM [keV]')
    ax.set_xlabel(r'$\tau$ [$\mu$s]')
    return fig

####################################################################################





"""
Offline analysis on waveforms obtained with dactylos
The shaping is done with code from 
https://github.com/alowell-ssl/awlpico
"""
import pylab as p
import numpy as np
import concurrent.futures as fut
import uproot as up
import tqdm
import os
import os.path
import random

import hepbasestack as hep 

from .utils import get_stripname, plot_waveform

# shaping stuff 
#from .shaping import GaussShaper, TrapezoidalFilter
from . import shaping as sh

import dactylos

logger = hep.logger.get_logger(dactylos.LOGLEVEL)

########################################################################

def read_waveform(infile, ch, entrystop=None, save_memory=True):
    """
    Return waveform data from a rootfile with uproot

    Args:
        infile (str)       : filename
        ch (int)           : digitizer channel

    Keyword Args:
        entrystop   (int)  : last entry - if None, read all
        save_memory (bool) : make sure to use as less memory as possible
                             by strictly sticking to int16
    Returns:
        ndarray : numpy array with waveform data. Waveform data is in
                  digitizer channels and of length of the record length
                  as set when configuring the digitizer
    """
    f = up.open(infile)
    data = f.get('ch' + str(ch)).get('waveform').array(entry_stop=entrystop)
    # we omit empty events here. This should not happen
    # but might happen during digitizer software debugging
    data = [baseline_correction(k, nsamples=1000)[0] for k in data if len(k) > 0]
    #if sh.WAVEFORMTYPE == 'ARRAY':
    #    data = np.array(data, dtype=np.int16)
    # data is in counts here, so it is still technichally not more precise than
    # an int 16
    #if sh.WAVEFORMTYPE == 'ARRAY':
    if save_memory:
        data = np.asarray(data, dtype=np.int16)
    logger.info(f'Read out {len(data)} events for channel {ch}')    
    return ch, data

########################################################################

def baseline_correction(input_pulses, nsamples=2000, save_memory=True):
    """
    Do a very simple baseline subtraction. Assume the first nsamples are
    the baseline, and then average over them and subtract it

    Args:
        input (ndarray)    : input waveform

    Keyword Args:
        nsamples (int)     : number of samples for baseline estimation 
                             at the beginning of the input data
        save_memory (bool) : make sure to use as less memory as possible
                             by strictly sticking to int16

    Returns:
        numpy.ndarray      : waveform with a subtracted baseline

    """

    absolute_zero = (2**14)/2
    baseline = input_pulses[:nsamples].mean()
    #print (baseline) 
    #baseline = input[:nsamples].mean() - zero
    #print (f"Baseline calculation gives us {baseline}")
    #return input - np.ones(len(input))*baseline - absolute_zero*np.ones(len(input)), baseline
    #if sh.WAVEFORMTYPE == 'ARRAY':
    if save_memory:
        data = input_pulses - np.ones(len(input_pulses), dtype=np.int16)*baseline, baseline
    else:
    #if sh.WAVEFORMTYPE == 'LIST':
        data = list(np.asarray(input_pulses, dtype=np.int16) - np.ones(len(input_pulses), dtype=np.int16)*int(baseline)), baseline
    return data
    #return baseline,(input - baseline)

########################################################################

class WaveformAnalysis(object):
    """
    Allow for the read-in and chaining of multiple files

    """

    def __init__(self, njobs=4,\
                 active_channels=[0,1,2,3,4,5,6,7],\
                 use_simple_trapezoid_shaper=False,\
                 shaper_order=4,\
                 adjust_shaper_order_dynamically=False):
        """
        Prepare the waveform analysis. This is a several step process.
        First, tell it what files to use, then load them and finaly analyze.

        Keyword Args:
            njobs (int)                         : number of cpu jobs to use
            use_simple_trapezoid_shaper (bool)  : instead of the Gaussian shaper, a simple trapezoidal
                                                  filter can be used. 

        Returns:
            order (int)           :  -- unknown shaping parameter
            decay_time (float)    :  The decay time of the pulse. This has to
                                     pretty much mach the decaying tail of the pulse.
                                     Decay time has to be given in seconds.
            dt (float)            :  Digitizer sample size (that is 1/sampling rate) in 
                                     seconds. This value is fix for the CAEN 6725 with 
                                     250 MSamples/s
            shaper_order (int)    :  Order of the Gaussian shaper (4 is default)
            adjust_shaper_order_dynamically (bool) : Switch to a higher shaper order
                                                     for larger peakingtimes automatically
        """
        self.files = []
        self.tpexecutor = fut.ThreadPoolExecutor(max_workers=njobs)
        self.ppexecutor = fut.ProcessPoolExecutor(max_workers=njobs)
        self.active_channels = active_channels
        self.channel_data = dict([(k, np.array([])) for k in self.active_channels])  
        self.order = shaper_order
        self.adjust_shaper_order_dynamically = adjust_shaper_order_dynamically
        self.use_simple_trapezoid_shaper = use_simple_trapezoid_shaper
        self.recordlengths = dict()

    def get_recordlengths(self):
        """
        For debugging purposes. Check the recordlength of the 
        Waveforms and return a set for each channel with the
        recordlengths in the actual waveform data. This might 
        differ from the actual setting for whatever reason. 
        Might be even a bug.

        Returns:
            dict     : ch -> set(recordlength)

        """
        result = dict()
        for ch in self.active_channels:
            rl = np.array([len(k) for k in self.channel_data[ch]])
            if len (set(rl)) != 1:
                raise ValueError(f'Waveforms in the SAME CHANNEL do not have the SAME recordlength! Something is terribly wrong! {set(rl)}')
            result[ch] = rl[0]
            
        self.recordlengths = result
        return result

    def get_eventcounts(self):
        """
        Return the number of events per channel. This only makes sense
        after some data has been via read_waveform

        Return:
            dict     : int->int Number of events per channel
        """
        nevents = dict()
        # readjust active channels
        # if one has 0 events, it is not active
        self.active_channels = []
        for ch in self.channel_data.keys():
            nevents[ch] = len(self.channel_data[ch])
            if nevents[ch] > 0:
                self.active_channels.append(ch)
            else:
                logger.warning(f'Ch {ch} has no observed events! Deactivating...')
        return nevents

    def __del__(self):
        # FIXME - for some reason this causes an exception
        #self.tpexecutor.shutdown()
        #self.ppexecutor.shutdown()
        pass

    def set_shaping_parameters(self, peaktime,
                               order = 4,\
                               decay_time = 80e-6,\
                               dt = 4e-9 ):
        """
        Set all the ncecessary parameters for Alex' gaussian
        shaper in a coherent way. 
        Typically order and decay_time do not need to be changed.

        Args: 
            peaktime (float)      :  The peaking time, that is the time
                                     where the waveform is sampled with 
                                     the gaussian. Peaktime has to be given in 
                                     seconds.

        Keyword Args:
            order (int)           :  -- unknown shaping parameter
            decay_time (float)    :  The decay time of the pulse. This has to
                                     pretty much mach the decaying tail of the pulse.
                                     Decay time has to be given in seconds.
            dt (float)            :  Digitizer sample size (that is 1/sampling rate) in 
                                     seconds. This value is fix for the CAEN 6725 with 
                                     250 MSamples/s
        """
        # peaktime can be e.g. 4e-6, that is 4 microseconds
        self.peaktime = peaktime
        self.order = order
        self.decay_time = decay_time
        self.dt = dt  

    def set_peakingtime_sequence(self, peakingtimes):
        """
        Set a number of peaking times we want to 
        try on each channel

        Args:
            peakingtimes (list)   : peakingtimes in microsec

        """
        self.peakingtime_sequence = peakingtimes

    def set_infiles(self, infiles):
        """
        Define a list of files to be readout later.

        Args:
            infiles (list)   : root files with waveform data
        
        """
        self.files = infiles

    def read_waveforms(self, entrystop=None, save_memory=True):
        """
        Read the waveforms from the files

        Keyword Args:
            entrystop (int)    : if not None, read only entrystop waveforms
            save_memory (bool) : make sure to use as less memory as possible
                                 by strictly sticking to int16

        Returns:
            None
        """
        future_to_fname = dict()

        for fname in self.files:
            # read out every file once per channel
            for ch in self.active_channels:
                #read_waveform(fname, ch)
                future_to_fname[self.tpexecutor.submit(read_waveform, fname, ch, entrystop = entrystop, save_memory=save_memory)] = fname

        logger.info("Readout jobs submitted..")
        init_ch = [False]*8
        for future in fut.as_completed(future_to_fname):
            ch, data = future.result()    
            #if sh.WAVEFORMTYPE == 'ARRAY':
            if save_memory:
                data = np.array([np.array(k, dtype=np.int16) for k in data])
            if not init_ch[ch]:
                self.channel_data[ch] = data
                init_ch[ch] = True
            else:
                self.channel_data[ch] = np.vstack((self.channel_data[ch], data))
        logger.info("Channel data created")
        return None
        #return channel_data

    def plot_waveforms(self, ch, nwaveforms,\
                       randomized=False,\
                       prefix='',\
                       savedir='.'):
        """
        Plot some of the waveforms.

        Args:
            ch (int)              : Plot waveforms of this channel
            nwaveforms (int)      : Plot the first nwaveforms
    
        Keyword Args: 
            randomized (bool)     : Instead of plotting the first <nwaveforms>
                                    plot <nwaverforms> random waveforms picked from anywhere 
                                    in the file.
            prefix     (str)      : A prefix to the filename
            savedir    (str)      : Save the resulting plot in this directory

        """
        wfplot = p.figure()
        # plot waveforms in a grid, always 3 colums
        lines = int(nwaveforms/3)
        if lines == 0 : lines = 1
        nwaveforms = lines*3

        if randomized: 
            indexes = [random.choice(np.arange(0,len(self.channel_data[ch]),1)) for k in range(nwaveforms)]
            logger.info(f'Will plot random waveforms with indexes {indexes}')
        #times = np.array([k for k in range(len(self.channel_data[ch][0]))])
        for i in tqdm.tqdm(range(nwaveforms), total=nwaveforms, desc="Plotting waveforms.."):
            ax = wfplot.add_subplot(lines, 3, i+1)
            if randomized:
                volts = self.channel_data[ch][indexes[i]]
            else:
                volts = self.channel_data[ch][i]
            # FIXME - temporary workaround
            times = np.array([k for k in range(len(volts))])
            ax = plot_waveform(ax, times, volts)     
            # plot the axes only on the leftmost plots
            # and the downmost
            if (i+1 - 3) == 0:
                ax.set_ylabel("d. chan.")
            if (i+1 - nwaveforms) < 4:
                ax.set_xlabel("time [$\mu$s]")
        stripname = get_stripname(ch) 
        if not prefix:
            savename = f'waveforms-{stripname}.png' 
        else:
            savename = f'waveforms-{prefix}-{stripname}.png'
        savename = os.path.join(savedir, savename)
        wfplot.savefig(savename)
        return savename 

    def analyze(self, channel, save_shp_file=False):
        """
        Applyt the gaussian shaping algorithm on the waveform data.
        
        Args:
            channel (int)        : Select the channel

        Keyword Args:
            save_shp_file (bool) : Save a file with the shaping times

        """
        ptime_energies = dict()
        if not self.recordlengths:
            self.get_recordlengths()

        #data = copy(self.channel_data[channel])
        data = self.channel_data[channel]
        for ptime in tqdm.tqdm(self.peakingtime_sequence, desc=f"Applying shaper for channel {channel}.."):
            if self.adjust_shaper_order_dynamically:
                if 500 < ptime <= 1000:
                    order = 3
                elif ptime <= 500:
                    order = 2
                else:
                    order = 7
            else:
                order = self.order
            if self.use_simple_trapezoid_shaper:
                #shaper = TrapezoidalFilter(ptime = ptime, recordlength = self.recordlengths[channel])
                shaper = sh.TrapezoidalFilter(ptime, 1000, self.recordlengths[channel])
            else:
                shaper = sh.GaussShaper(ptime, order=order)
            #if self.njobs > 1:
            energies = np.array([energy for energy in self.ppexecutor.map(shaper.shape_it, data, chunksize=10)], dtype=np.float16) # use 16 bit dtype, since the input is only 16 bit anyway.
                     # and this will save memory
            #else:
            #energies = np.array([shaper.shaper_it(wf) for wf in data])
            ptime_energies[ptime] = energies 

        #if self.njobs > 1:
        #    for future in fut.as_completed(future_to_fname):
        #        ch, data = future.result()    
            
        return ptime_energies 




"""
Models to fit the (x-ray) spectrum, finde the line position and
its associated resolution (fwhm).
"""
import numpy as np
import uproot as up
import os
import os.path
import time 

from copy import copy

import HErmes as he
import hepbasestack as hep
import dashi as d
import pylab as p

d.visual()

logger = hep.logger.get_logger(10)

from .utils import get_stripname

def model_with_shoulder_with_linear_cutoff(xs,\
                                           mu_p, fwhm_p, amp_p,\
                                           mu_s, fwhm_s, amp_s):
    """
    A full fledged model combining two gaussians - one for the Compton 
    shoulder and another one for the line peak. The shoulder has a linear
    cutoff in the reaion of the peak, ranging from mu_p - fwhm_p to mu_p. 
    for xs > mu_p the contribution of the shoulder is set to zero.

    Args:
        xs (array)       : input data
        mu_p (float)     : line peak
        fwhm_p (float)   : line resolution
        amp_p (float)    : number of events at the line peak
        mu_s (float)     : peak of the (Compton) shoulder
        fwhm_s (float)   : width of the shoulder
        amp_s (float)    : number of events at the peak of (Compton) 
                           shoulder

    Returns:
        ndarray          : model prediction
    """
    peak_ys = he.fitting.functions.fwhm_gauss(xs, mu_p, fwhm_p, amp_p) 
    shoulder_ys = he.fitting.functions.fwhm_gauss(xs, mu_s, fwhm_s, amp_s)
   
    linear_ys = np.zeros(len(peak_ys)) 
    # construct the cutoff-part
    if (np.isfinite(mu_p) and np.isfinite(fwhm_p)):
        start_x = xs[xs <= (mu_p - (abs(fwhm_p)/2))][-1]
        start_y = shoulder_ys[xs <= (mu_p - (abs(fwhm_p)/2))][-1]
        end_x   = mu_p
        assert start_x < end_x
        #end_y   = shoulder_ys[xs < mu_p][-1]
        end_y   = 0
        #print (f'linear interpolation from {start_x} to {end_x}. Yrange {start_y} to {end_y}')
        slope = (abs(start_y))/(abs(end_x) - abs(start_x))
        constant = slope*abs(end_x)

        #print (constant)
        for i, x in enumerate(xs):
            if (x >= start_x and x <= end_x):
                linear_ys[i] = -1*slope*x + constant
        #linear_interp = lambda xs : -1*slope*xs + np.ones(len(xs))*abs(start_y)
        #linear_ys[np.logical_and(xs >= start_x, xs<= end_x)] = linear_interp(xs[np.logical_and(xs >= start_x, xs<=end_x)])
        #print (linear_ys)
        #print (start_x, end_x, start_y)
        shoulder_ys[xs >= mu_p] = 0
    return peak_ys + shoulder_ys - linear_ys

########################################################################

def peak_only_model_with_linear_subtraction(xs, mu_p, fwhm_p, amp_p):
    """
    A model with a single peak, however, we subtract a linear
    contribution coming from a possible shoulder
    
    Args:
        xs (array)       : input data
        mu_p (float)     : line peak
        fwhm_p (float)   : line resolution
        amp_p (float)    : number of events at the line peak

    """
    peak_ys = he.fitting.functions.fwhm_gauss(xs, mu_p, fwhm_p, amp_p) 
    
    # construct the cutoff-part
    if (np.isfinite(mu_p) and np.isfinite(fwhm_p)):
        start_x = xs[xs < (mu_p - abs(fwhm_p)/2)][-1]
        start_y = peak_ys[xs < (mu_p - abs(fwhm_p)/2)][-1]
        end_x   = xs[xs < mu_p][-1]
        end_y   = peak_ys[xs < mu_p][-1]
        slope = (end_y - start_y)/(end_x - start_x)
        linear_interp = lambda xs : slope*xs + np.ones(len(xs))*start_y
        peak_ys[np.logical_and(xs >= start_x, xs<= end_x)] = linear_interp(xs[np.logical_and(xs >= start_x, xs<=end_x)])
        peak_ys[xs > end_x] = 0
    return peak_ys

##########################################################################

def create_model(energy, bins, startparams,\
            use_cutoff_for_shoulder=False,\
            limits = ((None, None),\
                      (None, None),\
                      (None, None),\
                      (None, None),\
                      (None, None),\
                      (None, None)),
            constrain_parameters=False,\
            mode='distribution'):
    """
    The fitting routine. Gaussian for peak + gaussian for Compton shoulder. As a default, energy 
    is all the energy values, and this function will create a distribution with binning as specified 
    in bins. This behaviour can be changed by modifying the mode switch.

    Args:
        energy (ndarray)               : energy as obtained by the digitizer
        bins (ndarray)                 : bins to create the distribution
        startparams (ndarray)          : a set of 6 startparams [mu_x, fwhm_x, amp_x, mu_x, fwhm_x, amp_x]

    Keyword Args:
        use_cutoff_for_shoulder (bool) : use a gaussian model with a cutoff
                                         for the fitting of the shoulder
        limits (tuple)                 : min/max limits for the fitparameters. Are
                                         passed through to 
                                         HErmes.fitting.Model.fit_to_data
        mode (str)                     : default is *distribution* which means that a histogram
                                         will be created out of *energy* with the binning *bins*.
                                         Alternatively, this can be set to *direct*, then *energy*
                                         needs to be bin content and bins need to be the corresponding
                                         bincenters. In this case there will no histogramming by 
                                         this function.
    """
    
    if use_cutoff_for_shoulder:
        # use a model with 2 gaussians and
        # a linear cutoff part at the 'end'
        # of the shoulder
        peakmod = he.fitting.Model(model_with_shoulder_with_linear_cutoff)
    else:
        # the x-ray peak
        peakmod = he.fitting.Model(he.fitting.functions.fwhm_gauss)
        # fit the compton shoulder with another gaussian to increase 
        # fit quality
        shouldermod = he.fitting.Model(he.fitting.functions.fwhm_gauss)
        peakmod += shouldermod
    
    if mode == 'distribution':
        peakmod.add_data(energy, bins=bins, create_distribution=True)
    elif mode == 'direct':
        assert len(energy) == len(bins), "In this mode, 'energy' and 'bins' must be the same length"
        peakmod.add_data(energy, xs=bins)
    else:
        raise ValueError(f'Mode {mode} not understood. Has to be one of "distribution" or "direct"')
    peakmod.startparams = startparams
    # we still constrain the parameters, 
    # condition is that xray peak 
    # and shoulder peak should not interfere
    peakmod.fit_to_data(silent=True,\
                        errors=(1,1,1,1,1,1),\
                        use_minuit=True,\
                        limits = limits)
    
    return peakmod

########################################################################

def first_guess(bincenters, bincontent, roi_right_edge=700):
    """
    Get a reasonable set of startparameters for a certain
    energy. This will look for the peak of the line and 
    then try to set the shoulde peak as far away as possible.

    Args:
        bincenters (ndarray)   : xs
        bincontent (ndarray)   : ys

    Keyword Args:
        roi_right_edge (int)   : the right edge of the roi (region-of-interest)

    """
    assert len(bincontent) == len(bincenters), f"Bincenters and bincontent must have the same length! {len(bincenters)} {len(bincontent)}"

    # we know that the spectrum is the full range of the digitizer, 
    # so first look at a region of interest.
    # typically, the peak will be around 500, 
    # so lets take 0 - 700 (default), then the peak 
    # should be savely in the right half of that part of the spectrum
    bincenters = bincenters[:roi_right_edge]
    bincontent = bincontent[:roi_right_edge]

    # assume we have a fair number of bins, so let's try 
    # to get the peak value for the line
    right_half = int(len(bincenters)/2)
    left_10_percent = int(right_half/5)
    logger.info(f'Found spectrum of len {len(bincenters)}. Dividing in half left and right of {right_half}')
    right_half_spectrum = bincontent[right_half:]
    peakval = np.where(right_half_spectrum == max(right_half_spectrum))[0][0]
    logger.info(f'Found an estimate for the line peak {peakval}')
    peakval += right_half 
    # estimate the fwhm
    peak_amp = max(right_half_spectrum)
    logger.info(f'Found an estimate for the peak amplitude {peak_amp}')
    mod = he.fitting.Model(he.fitting.fwhm_gauss)
    mod.startparams = (peakval, 5, peak_amp)
    mod.add_data(bincontent, xs=bincenters, create_distribution=False)
    mod.fit_to_data(silent=True,\
                    errors=(1,1,1),\
                    limits=((peakval - 0.01*peakval, peakval + 0.01*peakval),
                            (1, 100),
                            (peak_amp - 0.01*peak_amp, peak_amp + 0.01*peak_amp)))
    peak_fwhm = mod.best_fit_params[1]

    # then just ensure that the shoulder peak is far enough from the
    # line peak, maybe just start from the half?
    shoulder_peak = right_half
    # and finally assume the peak hight is 0.1 of the xray peak
    shoulder_amp = peak_amp/10
    
    # try to prefit it
    mod = he.fitting.Model(he.fitting.fwhm_gauss)
    mod.startparams = (shoulder_peak, 10, shoulder_amp) # 10 is arbitrary
    mod.add_data(bincontent, xs=bincenters, create_distribution=False)
    if peak_amp <=1:
        logger.warning("Estimate for the  peak amplitutde is doubtful...")
        peak_amp = 100 
    mod.fit_to_data(silent=True,\
                    errors=(1,1,1,1,1,1),\
                    limits=((left_10_percent, peakval - peak_fwhm),\
                            (1, 100),\
                            (1, peak_amp/2)))
    shoulder_fwhm = mod.best_fit_params[1]

    #first_guess_params = np.array([500,20,500, 300, 40, 50])
    first_guess_params = np.array([peakval, peak_fwhm, peak_amp,\
                                   shoulder_peak, shoulder_fwhm, shoulder_amp])
    limits = ((peakval-peak_fwhm, peakval+peak_fwhm),\
              (0.5*peak_fwhm, 2*peak_fwhm),\
              (1, peak_amp*1.1),\
              (left_10_percent, peakval-2*peak_fwhm),\
              (1, shoulder_fwhm*2),
              (1, peak_amp/2))
    for i,k in enumerate(limits):
        logger.debug(f"Setting limit on parameter {i} of {k}")
    return first_guess_params, limits    


########################################################################

def fit_file(infilename = '143_4000.root',\
             file_type = '.root',\
             channel = 0,\
             ptime = -1,\
             savename_prefix = "",\
             detid = -1,\
             plot_dir = '.',\
             energy = None,\
             #bins = np.linspace(150, 550, 200),\
             bins = np.linspace(1,2**14, 2**14) - 0.5,\
             peakposition = 88.03,\
             debug=False):
    """
    One shot function to fit a datafile with 
    energy values obtained by the Caen N6725 digitizer
    with two gaussians. One for the signal peak,
    another one for the Compton shoulder.

    Keyword Args:
        infilename (str)     : the file to fit. Must have energy  
        file_type (str)      : A specialized string to identify the file type
                               Can be either '.root' (default),'.txt' or None
                               In case of '.txt' one line per digitizer 
                               bin is expected (2**14 lines) with one
                               value each. If None is given, then the parameter energy
                               has to be different than None. In that case,
                               there is no need to read out a file 
        channel (int)        : digitizer channel (0-7)
        energy  (ndarray)    : in case there is no file, but we 
                               have the data already, it can be passed via
                               the energies parameter. If there is a file, 
                               this has to be 'None'
        plot_dir (int)       : directory to save the plots in 
        savename_prefix (str): a prefix to be added to the png file the plot is saved as
        bins (ndarray)       : bins for the histogram
        ptime (float)        : peaking time (just used for the plot title)
        detid (int)          : detector id (just uded for the plot title)
        peakposition (float) : the true value of the x-ray peak in keV
                               this is used to recalibrate the x-axis
        debug        (bool)  : Use to reduce the loglevel, otherwise this 
                               function is as silent as possible.
                               Internally sets the loglevel of HErmes to debug
    """
    if not debug:
        he.set_loglevel(100) 
        logger.setLevel(100)

    # set startparameters
    startparams = np.array([500,20,500, 300, 40, 50])
    metainfo = "" # this currently holds the type of measurement
                  # if it is from a file -> CAEN shaping
                  # mode direct -> mc2 
                  # else -> waveform
    if file_type == '.root':
        mode = 'distribution'
        f = up.open(infilename)
        trigger = f.get('ch' + str(channel)).get('trigger').array()
        energy  = f.get('ch' + str(channel)).get('energy').array()    
        # prebin the histogram to estimate start params
        h = d.factory.hist1d(energy, bins)
        startparams, limits = first_guess(h.bincenters, h.bincontent)
        peakmod = create_model(energy, bins, startparams, limits=limits)
        metainfo = 'TrpzShap'
    elif file_type == '.txt':
        mode='direct'
        energy = []
        with open(infilename) as f:
            for line in f.readlines():
                binenergy = 0
                if len(line.split()) == 2:
                    # we have 2 column data. second line is the bincontent
                    # FIXM: for some reason, in the mc2 data, the first
                    # two bins are a float. This can be bascially ignored
                    # since it is out of our region of interest
                    binenergy = int(float(line.split()[1]))
                else:
                    # one column data, just energy
                    binenergy = int(line.split()[0])
                energy.append(binenergy) 
        metainfo = 'TrpzShapMC2'

        energy = np.array(energy)
        if len(energy) != 2**14:
            raise ValueError(f"Bins missing! Only found {len(energy)} bins in the file!")
        bins = np.linspace(1,2**14, 2**14)
        # make them the bincenter instead of the edge
        bins = bins - 0.50
        startparams, limits = first_guess(bins, energy)
        peakmod = create_model(energy, bins, startparams, mode=mode, limits=limits)

    elif file_type == None:
        if energy is None:
            raise ValueError("Need to give energy or filename and filetype!")
        # direct mode, no file. in that energy and bins must be given directly
        mode = 'distribution'
        # prebin the histogram to estimate start params
        h = d.factory.hist1d(energy, bins)
        logger.info(f'Caluclating start parameters')
        startparams, limits = first_guess(h.bincenters, h.bincontent)
        peakmod = create_model(energy, bins, startparams, mode=mode, limits=limits)
        metainfo = 'GaussShap' 
    else:
        raise ValueError(f'Can not process file type {file_type}')

    logger.info(f'Got chi2/ndf of {peakmod.chi2_ndf}')
    logger.info(f'Found peak at {peakmod.best_fit_params[0]}')
    logger.info(f'Best fit parameters {peakmod.best_fit_params}')

    peak = peakmod.best_fit_params[0]

    # recalibrate the x-axis
    logger.info(f'Calibrating x-axis...{peak} -> {peakposition}')
    conversion = peakposition/peak

    # only convert the energies if they 
    # are truly energies.
    # in the 'direct' approach for mc2 data
    # we 'hijack' these to be the bincontents
    if mode =='direct':
        energyxs = energy
    else:
        energyxs = conversion*energy
    

    # use the previously obtained fit params as start params
    # remember the amplitdue is not converted.
    startparams = [conversion*startparams[0], conversion*startparams[1],\
                   startparams[2], conversion*startparams[3],\
                   conversion*startparams[4], startparams[5]]

    limits = [conversion*np.array(limits[0]),\
              conversion*np.array(limits[1]),\
              np.array(limits[2]),\
              conversion*np.array(limits[3]),\
              conversion*np.array(limits[4]),\
              np.array(limits[5])]
    limits = tuple(limits)
    bins = conversion*bins

    startparams = np.array([abs(s) for s in startparams])
    # get a better fit with a more refined model
    logger.info(f'Refitting...with startparams {startparams}')
    peakmod = create_model(energyxs, bins, startparams,\
                      limits = limits,\
                      use_cutoff_for_shoulder= False,\
                      mode=mode)
    logger.info(f'Got chi2/ndf of {peakmod.chi2_ndf}')
    logger.info(f'Found peak at {peakmod.best_fit_params[0]} with a FWHM of {peakmod.best_fit_params[1]}')

    # create a nice figure
    #fig = p.figure(figsize=SINGLE_FIG, dpi=DPI)
    fig = p.figure()
    #fig = matplotlib.figure.Figure()
    ax = fig.gca()
    if mode == 'direct':
        ax.scatter(peakmod.xs, peakmod.data, c="k",\
                   edgecolors="k", marker="+",\
                   s=20, linewidth=1  )
    else:
        peakmod.distribution.scatter()
    ax.plot(peakmod.xs, peakmod.prediction(peakmod.xs), color="r")
    ax = hep.visual.adjust_minor_ticks(ax)
    ax.set_xlabel('energy [keV]')
    ax.set_ylabel('events')
    
    # error handling
    errdict = peakmod.errors

    # compose a title
    stripname = get_stripname(channel)
    title = f'det. {detid} {stripname} pt {ptime/1000}$\mu$s'
    ax.set_title(title, loc='right')

    # create a textbox with some output
    infotext  = r"\begin{tabular}{ll}"
    infotext += r"spectra fit:&\\ "
    if mode == 'direct':
        infotext += r"entries: & {:4.2f} \\".format(sum(peakmod.data))
    else:
        infotext += r"entries: & {:4.2f} \\".format(peakmod.distribution.stats.nentries)
    infotext += r"peak (line)  & {:4.2f} $\pm$ {:4.2f} \\ ".format(peakmod.best_fit_params[0], errdict['mu0'])
    infotext += r"FWHM (line)  & {:4.2f} $\pm$ {:4.2f} \\ ".format(peakmod.best_fit_params[1], errdict['fwhm0'])
    if len(peakmod.best_fit_params) > 3:
        infotext += r"peak (shoulder) & {:4.2f} $\pm$ {:4.2f} \\ ".format(peakmod.best_fit_params[3], errdict['mu1'])
        infotext += r"FWHM (shoulder) & {:4.2f} $\pm$ {:4.2f} \\ ".format(peakmod.best_fit_params[4], errdict['fwhm1'])
        infotext += r"$\chi^2/ndf$ & {:4.2f} \\ ".format(peakmod.chi2_ndf)
    infotext += r"\end{tabular}"
    ax.text(0.3, 0.5, infotext,\
            horizontalalignment='center',\
            verticalalignment='center',\
            transform=ax.transAxes,\
            bbox=dict(facecolor='white', alpha=0.7, edgecolor=None),\
           )

    # in case we have one bin per digitizer 
    # channel, cut down on the xrange
    if len(bins) > 1000:
        # ignore the first 10 mV
        # we will not see a peak at zero,
        # but actually we don't care
        # and this helps with the scaling
        ax.set_xlim(left=0, right=110)
        if (ax.get_ylim()[1] > 2*max(energyxs[20:])):
            ax.set_ylim(top= 1.1*max(energyxs[20:]),
                        bottom = 0)

    # return the resolution
    pngfilename = f'det{detid}-stime{ptime}-{stripname}-{metainfo}.png' 
    if savename_prefix:
        pngfilen = savename_prefix + pngfilename
    logger.info(f'Saving {pngfilename} in {plot_dir}... ')
    fig.savefig(os.path.join(plot_dir,pngfilename))
    time.sleep(1) # give the filesystem time to save the figure
                  # not sure why it seems that there are some hickups
    p.close(fig)
    logger.info('..done')
    # copy them to decouple them from the model
    # since some minuit entities are non-picklable
    # which prevents this from running multithreaded
    return_pars = copy(peakmod.best_fit_params)
    return_errs = dict()
    for k in errdict.keys():
        return_errs[k] = errdict[k]
    #@return detid, channel, peakmod.best_fit_params[1], errdict['fwhm0'], pngfilename
    return detid, channel, return_pars, return_errs, pngfilename




"""
setup.py file for dactylos - CAEN N6725 digitizer read out software.
This setup.py file will compile the C++ extensions needed to interact
with the digitizer with the help of an additional CMakeLists.txt file.
"""

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

import os.path, re, sys
#import pybind11

import os
import re
import sys
import sysconfig
import platform
import subprocess
import shlex
from pathlib import Path
from glob import glob

from distutils.version import LooseVersion
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.test import test as TestCommand

def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with open(os.path.join(package, '__init__.py'), 'rb') as init_py:
        src = init_py.read().decode('utf-8')
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", src).group(1)

VERSION = get_version('dactylos')

tests_require = [
    'pytest>=3.0.5',
    'pytest-cov',
    'pytest-runner',
]

needs_pytest = set(('pytest', 'test', 'ptr')).intersection(sys.argv)
setup_requires = ['pytest-runner'] if needs_pytest else []
setup_requires.append('pybind11>2.4')

#####################################################
# Get dependnecies 
# ROOT lib/include dirs

def get_root_include_dir():
    rootsys = os.getenv('ROOTSYS')
    if rootsys is None:
        raise SystemError("$ROOTSYS shell variable not defined! Make sure to have root installed end this variable defined.")
    print (f'Found root include dir at {rootsys}/include')
    return os.path.join(rootsys, 'include')

def get_root_lib_dir():
    rootsys = os.getenv('ROOTSYS')
    if rootsys is None:
        raise SystemError("$ROOTSYS shell variable not defined! Make sure to have root installed end this variable defined.")
    print (f'Found root include dir at {rootsys}/lib')
    return os.path.join(rootsys, 'lib')

class get_pybind_include(object):
    """
    Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. 
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def get_root_cxx_standard():
    """
    Invoce root-config to find out which cxx standard was used
    to compile root
    """
    command = shlex.split('root-config --cflags')
    result = subprocess.Popen(command, stdout=subprocess.PIPE).communicate()[0]
    cxxstandard = re.compile('-std=c\+\+(?P<std>[0-9]+)')
    try:
        cxxstandard = cxxstandard.search(result.decode()).groupdict()['std']
    except Exception as e:
        print(f'Can not retrieve uxed C++ standard used when compiling root, exceoption {e}')
    return cxxstandard


# since root might be not in the superusers python path, add it to 
# sys path
sys.path.append(get_root_lib_dir())
import ROOT 

#####################################################

# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """
    Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True

#####################################################

class CMakeExtension(Extension):
    """
    Allows to install extensions with the
    help of an external CMakeLists.txt file
    """

    def __init__(self, name, **kwargs):
        Extension.__init__(self, name, **kwargs)
        print (f'Found extension with neame {name}')
        if name.startswith('_py'):
            self._cfilename = name# + '.so'
        elif name.startswith('_tr'):
            self._cfilename = name# + '.so'
        else:
            self._cfilename = 'lib' + name# + '.so'
        print (f'library name {self._cfilename}')
        self.has_explicit_destination = False


class CMakeBuild(build_ext):
    """
    Invoke cmake and an external CMakeLists.txt file to 
    build the external components

    """

    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        build_directory = os.path.abspath(self.build_temp)

        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + build_directory,
            #'-DPYTHON_EXECUTABLE=' + sys.executable
            #'-DPYTHON_EXECUTABLE=/usr/bin/python3', \
            #'-DPYTHON_INCLUDE_DIR=/usr/include/python3.6m', \
            #'-DPYTHON_LIBRARY=/usr/lib/x86_64-linux-gnu/libpython3.6m.so'
            f'-DCMAKE_CXX_STANDARD={get_root_cxx_standard()}'\
        ]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        # Assuming Makefiles
        build_args += ['--', '-j2']

        self.build_args = build_args

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # CMakeLists.txt is in the same directory as this setup.py file
        cmake_list_dir = os.path.abspath(os.path.dirname(__file__))
        print('-'*10, 'Running CMake prepare', '-'*40)
        subprocess.check_call(['cmake', cmake_list_dir] + cmake_args,
                              cwd=self.build_temp, env=env)

        print('-'*10, 'Building extensions', '-'*40)
        cmake_cmd = ['cmake', '--build', '.'] + self.build_args
        subprocess.check_call(cmake_cmd,
                              cwd=self.build_temp)

        # Move from build temp to final position
        #print (self.extensions)
        

        for ext in self.extensions:
            build_temp = Path(self.build_temp).resolve(ext.name)
            print (f'Found build dir {build_temp}')        
            source_file = os.path.join(build_temp, ext._cfilename)
            print (f'Looking for source file {source_file}' + '*.so')
            source_files = glob(source_file + "*")
            if not source_files:
                print (f'WARN: {source_file}*.so not found...!') 
                continue
                #raise SystemError("Can not find source file!")
            source_file = source_files[0]
            # destination is the directory in the install path (hopefully)
            dest_path = os.path.split(Path(self.get_ext_fullpath(ext._cfilename)).resolve())[0] 
            dest_path = Path(dest_path) / 'dactylos' 
            if ext._cfilename.startswith('_tr'):
                # put trapezoidal shaper library in analysis
                dest_path = dest_path / 'analysis' / 'shaping'
            print (f'.. will copy to {dest_path}')
            print (f"Trying to copy {ext.name} from {source_file} to {dest_path}")
            self.copy_file(source_file, dest_path)
        # copy the dactylos library
            

# external modules, build by CMake. At the moment this is all 
# double a little bit, this must be also defined in the CMakeList.txt file
# this just helps for the actual install process
ext_modules = [
    CMakeExtension(
        'Dactylos',
        sources = ['src/trapezoidal_shaper.cxx',
                   'src/CaenN6725.cxx'],
        include_dirs=[
            # Path to pybind11 headers
            #get_pybind_include(),
            #get_pybind_include(user=True),
            "include",
            get_root_include_dir()
        ],
        libraries=['Dactylos'],
        language='c++'
    ),
    CMakeExtension(
        '_pyCaenN6725',
        sources = ['src/module.cxx'],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            "include",
            get_root_include_dir()
        ],
        libraries=['CAENDigitizer','CaenN6725'],
        language='c++'
    )
]



setup(name='dactylos',
      version=VERSION,
      description='Python package to interact and readout CAEN N6725 digitizers. Can access waveform information, energy values of the shapers and allows for easy configuration of the instrument.',
      long_description='This is a private project, and without association of CAEN in any kind. There is no guarantee that this code is useful or working or not harmful. Please see the licensce agreement. This code needs the CAEN C libraries for communication of the digitizer via USB as well as the CAEN C interface library, see https://www.caen.it/products/n6725/.',
      author='Achim Stoessl',
      author_email="achim.stoessl@gmail.com",
      url='https://github.com/achim1/dactylos',
      #download_url="pip install skippylab",
      install_requires=['numpy>=1.11.0',
                        'matplotlib>=1.5.0',
                        'six>=1.1.0',
                        'pybind11>2.4'],
    ext_modules=ext_modules,
    #cmdclass={'build_ext': BuildExt},
    zip_safe=False,
    setup_requires=setup_requires,
    #tests_require=tests_require,
    license="GPL",
    platforms=["Ubuntu 18.04"],
    cmdclass=dict(build_ext=CMakeBuild),
    classifiers=[
      "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Science/Research",
      "Intended Audience :: Developers",
      "Programming Language :: Python :: 3.6",
      "Topic :: Scientific/Engineering :: Physics"
            ],
    keywords=["digitzer", "CAEN",\
              "CAEN N6725", "6725",\
              "readout", "physics", "engineering", "lab", "USB"],
    packages=['dactylos', 'dactylos.analysis', 'dactylos.analysis.shaping'],
    # FIXME: put the header for the trapezoidal shpaer somewhere else
    #data_files = [('.', ['CMakeLists.txt']),
    #             ],
                 # ('dactylos/analysis/shaping/', ['trapezoidal_shaper.h'])],
    # use the package_data hook to get the pybindings (as compiled with cmake) to the right
    # final destination
    #package_data={'dactylos.analysis.shaping' : ['*.so']},
    #include_package_data=True,
    scripts=['bin/RunDigitizer', 'bin/FitXrayData']
    )





from .gauss_shaper import GaussShaper

USE_CXX_EXTENSION=True

if USE_CXX_EXTENSION:
    try:
        #from _trapezoidal_shaper import TrapezoidalFilter
        from dactylos._pyCaenN6725 import TrapezoidalFilter
    except ImportError as e:
        print (f"WARNING, can not import c++  extension for TrapezoidalFilter, Trapezoidal filter will be SLOW!. Exception {e}")
        from .trapezoidal_shaper import TrapezoidalFilter
else:
    from .trapezoidal_shaper import TrapezoidalFilter
    print (f"WARNING, not using c++  extension for TrapezoidalFilter, Trapezoidal filter will be SLOW!.")




"""
Analysis tools for data taken with dactylos.CaenN6725 digitizer software
"""

import numpy as _np
import pylab as _p
import hepbasestack as _hep

from .file_inspector import inspect_file
from .noisemodel import fit_noisemodel
from .utils import get_stripname

##########################################################

def fit_all_strips(data, detector_id,\
                   use_direct_model=False,\
                   use_minuit=False,\
                   debug_minuit=False,\
                   method='mc2',\
                   plotdir='.'):
    """
    Create the noisemodel fits for all strips for a certain detector

    Args:
        data ()                  : data obtained by reading in a summary file
        detector_id        (int) : unique detector identifier #
    Keyword Args:
        use_direct_model  (bool) : use a different way of fitting the noisemodel.
                                   Fit the noiseparameters directly
        use_minuit        (bool) : use iminuit for the fit.
                                   Alternative is scipy.optimize.lseastsq
        debug_minuit      (bool) : attach more informations to the noisemodel which 
                                   can help to resolve fitting isues
        method             (str) : the method which was used to create the data in the
                                   summaryfile
        plotdir            (str) : save plots in this directory
    """

    nmodels = []
    for i,strip in enumerate([f'{get_stripname(k)}' for k in range(8)]):
        ptime    = _np.asarray([k.ptime for k in data if k.strip == strip])
        fwhm     = _np.asarray([k.fwhm  for k in data if k.strip == strip])
        fwhm_err = _np.asarray([k.fwhme for k in data if k.strip == strip])
        fit_method = 'trapezoid'
        if data[0].mtype == 'WF':
            fit_method = 'gauss'
        try:
            fig, model =    fit_noisemodel(ptime, fwhm, fwhm_err,\
                                           i, detector_id, plotdir=plotdir,\
                                           use_minuit=use_minuit,
                                           use_direct_model=use_direct_model,\
                                           debug_minuit=debug_minuit,
                                           fig=_p.figure(figsize=\
                                                    _hep.layout.FIGSIZE_A4_LANDSCAPE,\
                                                    dpi=120),\
                                           method=fit_method
                                          )

        except Exception as e:
            print (f'Can not fit strip {i}, exception {e}')
            continue
        #fig.savefig(os.path.join(plotdir,f'nm-{detector_id}-{strip}-{method}.png'))
        nmodels.append(model)
        #break
    return fig, nmodels




import uproot as up
from dataclasses import dataclass

ENERGY = 'energy'
WAVEFORM = 'waveform'
TRIGGER = 'trigger' 

@dataclass
class ChannelInspector:
    """
    Summary information about a specific channel
    """
    channel_id  : int = 0
    n_energies  : int = 0
    n_waveforms : int = 0
    n_trigger   : int = 0

def inspect_file(filename):
    """
    Gather basic information from a file. Active channels, 
    events per channel etc.

    Args;
        filename (str) : The file to be instpected
    """

    inspector  = dict()
    ch_inspector = dict()
    f = up.open(filename)
    for k in range(8):
        chdata = f.get(f'ch{k}')
        ch_inspect = ChannelInspector()
        ch_inspect.channel_id  = k
        ch_inspect.n_energies  = 0
        ch_inspect.n_waveforms = 0 
        ch_inspect.n_trigger   = 0
        try:
            ch_inspect.n_energies  = chdata.get(ENERGY).num_entries
        except Exception as e:
            print (f'Can not get energies, exception {e}')
        try:
            ch_inspect.n_waveforms = chdata.get(WAVEFORM).num_entries
        except Exception as e:
            print (f'Can not get waveforms, exception {e}')
        try:
            ch_inspect.n_trigger   = chdata.get(TRIGGER).num_entries
        except Exception as e:
            print (f'Can not get triggers, exception {e}')
        inspector[k] = ch_inspect

    return inspector
